from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as K


class WeightDecayScheduler(Callback):
    def __init__(self, warmup_schedule, decay_schedule, verbose=0):
        super().__init__()
        self.warmup_schedule = warmup_schedule
        self.decay_schedule = decay_schedule
        self.epoch = 0
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch = epoch
        if not hasattr(self.model.optimizer, "weight_decay"):
            raise ValueError('Optimizer must have a "weight_decay" attribute.')
        wd = float(K.get_value(self.model.optimizer.weight_decay))
        new_wd = self.decay_schedule(epoch, wd)
        if new_wd == wd:
            return
        K.set_value(self.model.optimizer.weight_decay, K.get_value(new_wd))
        if self.verbose > 0:
            print(
                "\nEpoch %05d: WeightDecayScheduler reducing weight decay "
                "to %s." % (epoch + 1, new_wd)
            )

    def on_train_batch_begin(self, batch, logs=None):
        if not hasattr(self.model.optimizer, "weight_decay"):
            raise ValueError('Optimizer must have a "weight_decay" attribute.')
        wd = float(K.get_value(self.model.optimizer.weight_decay))
        step = self.epoch * self.params["steps"] + batch
        new_wd = self.warmup_schedule(step, wd)
        if new_wd == wd:
            return
        K.set_value(self.model.optimizer.weight_decay, K.get_value(new_wd))
        if self.verbose > 0:
            print(
                "\nStep %05d: WeightDecayScheduler increasing weight decay "
                "to %s." % (step + 1, new_wd)
            )

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs["weight_decay"] = K.get_value(self.model.optimizer.weight_decay)



from typing import Optional

import tensorflow as tf
from tensorflow.keras.layers import Layer, Bidirectional, Dropout

from sknlp.typing import WeightRegularizer, WeightInitializer, WeightConstraint

from .lstmp import LSTMP


class MultiLSTMP(Layer):
    def __init__(
        self,
        num_layers: int,
        units: int = 200,
        projection_size: int = 100,
        kernel_initializer: WeightInitializer = "glorot_uniform",
        recurrent_initializer: WeightInitializer = "orthogonal",
        projection_initializer: WeightInitializer = "glorot_uniform",
        bias_initializer: WeightInitializer = "zeros",
        kernel_regularizer: Optional[WeightRegularizer] = None,
        recurrent_regularizer: Optional[WeightRegularizer] = None,
        projection_regularizer: Optional[WeightRegularizer] = None,
        bias_regularizer: Optional[WeightRegularizer] = None,
        kernel_constraint: Optional[WeightConstraint] = None,
        recurrent_constraint: Optional[WeightConstraint] = None,
        projection_constraint: Optional[WeightConstraint] = None,
        bias_constraint: Optional[WeightConstraint] = None,
        input_dropout: float = 0.0,
        recurrent_dropout: float = 0.0,
        output_dropout: float = 0.0,
        recurrent_clip: Optional[float] = None,
        projection_clip: Optional[float] = None,
        last_connection: str = "last",
        **kwargs
    ):
        self.num_layers = num_layers
        self.last_connection = last_connection
        self.output_dropout = output_dropout

        self.layers = []
        for i in range(num_layers):
            return_sequences = i != num_layers - 1 or last_connection != "last"
            self.layers.append(
                Bidirectional(
                    LSTMP(
                        units,
                        projection_size,
                        kernel_initializer=kernel_initializer,
                        recurrent_initializer=recurrent_initializer,
                        projection_initializer=projection_initializer,
                        bias_initializer=bias_initializer,
                        kernel_regularizer=kernel_regularizer,
                        recurrent_regularizer=recurrent_regularizer,
                        projection_regularizer=projection_regularizer,
                        bias_regularizer=bias_regularizer,
                        kernel_constraint=kernel_constraint,
                        recurrent_constraint=recurrent_constraint,
                        projection_constraint=projection_constraint,
                        bias_constraint=bias_constraint,
                        dropout=input_dropout,
                        recurrent_dropout=recurrent_dropout,
                        recurrent_clip=recurrent_clip,
                        projection_clip=projection_clip,
                        return_sequences=return_sequences,
                    )
                )
            )
        super().__init__(**kwargs)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        for layer in self.layers:
            inputs = layer(inputs)
        if self.output_dropout:
            noise_shape = (None, 1, self.layers[0].forward_layer.projection_size * 2)
            if self.last_connection == "last":
                noise_shape = None
            return Dropout(
                self.output_dropout, noise_shape=noise_shape, name="output_dropout"
            )(inputs)
        else:
            return inputs

    def get_config(self):
        return {
            **super().get_config(),
            "num_layers": self.num_layers,
            "units": self.layers[0].forward_layer.units,
            "projection_size": self.layers[0].forward_layer.projection_size,
            "kernel_initializer": self.layers[0].forward_layer.kernel_initializer,
            "recurrent_initializer": self.layers[0].forward_layer.recurrent_initializer,
            "projection_initializer": self.layers[
                0
            ].forward_layer.projection_initializer,
            "bias_initializer": self.layers[0].forward_layer.bias_initializer,
            "kernel_regularizer": self.layers[0].forward_layer.kernel_regularizer,
            "recurrent_regularizer": self.layers[0].forward_layer.recurrent_regularizer,
            "projection_regularizer": self.layers[
                0
            ].forward_layer.projection_regularizer,
            "bias_regularizer": self.layers[0].forward_layer.bias_regularizer,
            "kernel_constraint": self.layers[0].forward_layer.kernel_constraint,
            "recurrent_constraint": self.layers[0].forward_layer.recurrent_constraint,
            "projection_constraint": self.layers[0].forward_layer.projection_constraint,
            "bias_constraint": self.layers[0].forward_layer.bias_constraint,
            "input_dropout": self.layers[0].forward_layer.dropout,
            "recurrent_dropout": self.layers[0].forward_layer.recurrent_dropout,
            "recurrent_clip": self.layers[0].forward_layer.recurrent_clip,
            "projection_clip": self.layers[0].forward_layer.projection_clip,
            "output_dropout": self.output_dropout,
            "last_connection": self.last_connection,
        }



from __future__ import annotations
from typing import Callable, Optional, Any
import logging
import tensorflow as tf

from sknlp.utils.tagging import (
    viterbi_decode,
    convert_ids_to_tags,
    tagging_fscore,
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class TaggingFScoreMetric(tf.keras.callbacks.Callback):
    def __init__(
        self,
        idx2tag: Callable[[int], str],
        labels: list[str],
        pad_tag: str,
        start_tag: Optional[str] = None,
        end_tag: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.idx2tag = idx2tag
        self.labels = labels
        self.pad_tag = pad_tag
        self.start_tag = start_tag
        self.end_tag = end_tag

    def on_epoch_end(self, epoch: int, logs: Optional[dict[str, Any]] = None) -> None:
        if self.validation_data is None:
            return

        emissions, mask = self.model.predict(self.validation_data)
        tag_ids_list = viterbi_decode(
            self.model.get_layer("crf").get_weights()[0],
            emissions.transpose(1, 0, 2),
            mask.transpose(1, 0),
        )
        predictions = []
        for tag_ids in tag_ids_list:
            predictions.append(
                convert_ids_to_tags(
                    self.idx2tag,
                    tag_ids,
                    self.pad_tag,
                    start_tag=self.start_tag,
                    end_tag=self.end_tag,
                )
            )

        y = []
        for _, tag_ids_array in self.validation_data.as_numpy_iterator():
            y.extend(
                [
                    convert_ids_to_tags(
                        self.idx2tag,
                        tag_ids,
                        self.pad_tag,
                        start_tag=self.start_tag,
                        end_tag=self.end_tag,
                    )
                    for tag_ids in tag_ids_array.tolist()
                ]
            )

        score_df = tagging_fscore(y, predictions, self.labels)
        logger.info(score_df)
        row = score_df[score_df["class"] == "avg"]
        for col in score_df.columns:
            if col == "support":
                continue
            logs[f"val_{col}"] = row[col].values.tolist()[0]


from typing import List, Dict, Any

import tensorflow as tf

from sknlp.layers import MultiLSTMP, MLPLayer, LSTMP, LSTMPCell, CrfLossLayer
from .deep_tagger import DeepTagger


class TextRNNTagger(DeepTagger):
    def __init__(
        self,
        classes: List[str],
        max_sequence_length: int = 100,
        sequence_length: int = None,
        segmenter: str = "char",
        embedding_size: int = 100,
        use_batch_normalization: bool = True,
        num_rnn_layers: int = 1,
        rnn_hidden_size: int = 512,
        rnn_projection_size: int = 128,
        rnn_recurrent_clip: float = 3.0,
        rnn_projection_clip: float = 3.0,
        rnn_input_dropout: float = 0.5,
        rnn_recurrent_dropout: float = 0.5,
        rnn_output_dropout: float = 0.5,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 128,
        fc_activation: str = "tanh",
        fc_momentum: float = 0.9,
        fc_epsilon: float = 1e-5,
        rnn_kernel_initializer="glorot_uniform",
        rnn_recurrent_initializer="orthogonal",
        rnn_projection_initializer="glorot_uniform",
        rnn_bias_initializer="zeros",
        rnn_kernel_regularizer=None,
        rnn_recurrent_regularizer=None,
        rnn_projection_regularizer=None,
        rnn_bias_regularizer=None,
        rnn_kernel_constraint=None,
        rnn_recurrent_constraint=None,
        rnn_projection_constraint=None,
        rnn_bias_constraint=None,
        text2vec=None,
        **kwargs
    ):
        super().__init__(
            classes,
            start_tag=None,
            end_tag=None,
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=segmenter,
            embedding_size=embedding_size,
            use_batch_normalization=use_batch_normalization,
            text2vec=text2vec,
            algorithm="rnn",
            **kwargs
        )
        self.num_rnn_layers = num_rnn_layers
        self.rnn_hidden_size = rnn_hidden_size
        self.rnn_projection_size = rnn_projection_size
        self.rnn_recurrent_clip = rnn_recurrent_clip
        self.rnn_projection_clip = rnn_projection_clip
        self.rnn_input_dropout = rnn_input_dropout
        self.rnn_recurrent_dropout = rnn_recurrent_dropout
        self.rnn_output_dropout = rnn_output_dropout
        self.rnn_kernel_initializer = rnn_kernel_initializer
        self.rnn_recurrent_initializer = rnn_recurrent_initializer
        self.rnn_projection_initializer = rnn_projection_initializer
        self.rnn_bias_initializer = rnn_bias_initializer
        self.rnn_kernel_regularizer = rnn_kernel_regularizer
        self.rnn_recurrent_regularizer = rnn_recurrent_regularizer
        self.rnn_projection_regularizer = rnn_projection_regularizer
        self.rnn_bias_regularizer = rnn_bias_regularizer
        self.rnn_kernel_constraint = rnn_kernel_constraint
        self.rnn_recurrent_constraint = rnn_recurrent_constraint
        self.rnn_projection_constraint = rnn_projection_constraint
        self.rnn_bias_constraint = rnn_bias_constraint
        self.num_fc_layers = num_fc_layers
        self.fc_hidden_size = fc_hidden_size
        self.fc_activation = fc_activation
        self.fc_momentum = fc_momentum
        self.fc_epsilon = fc_epsilon
        self.inputs = [
            tf.keras.Input(shape=(None,), dtype=tf.int32, name="token_id"),
            tf.keras.Input(shape=(None,), dtype=tf.int32, name="tag_id"),
        ]

    def build_encode_layer(self, inputs):
        token_ids, tag_ids = inputs
        mask = tf.keras.layers.Lambda(
            lambda x: tf.cast(x != 0, tf.int32), name="mask_layer"
        )(token_ids)
        return (
            MultiLSTMP(
                self.num_rnn_layers,
                self.rnn_hidden_size,
                projection_size=self.rnn_projection_size,
                recurrent_clip=self.rnn_recurrent_clip,
                projection_clip=self.rnn_projection_clip,
                input_dropout=self.rnn_input_dropout,
                recurrent_dropout=self.rnn_recurrent_dropout,
                output_dropout=self.rnn_output_dropout,
                last_connection="all",
                kernel_initializer=self.rnn_kernel_initializer,
                recurrent_initializer=self.rnn_recurrent_initializer,
                projection_initializer=self.rnn_projection_initializer,
                bias_initializer=self.rnn_bias_initializer,
                kernel_regularizer=self.rnn_kernel_regularizer,
                recurrent_regularizer=self.rnn_recurrent_regularizer,
                projection_regularizer=self.rnn_projection_regularizer,
                bias_regularizer=self.rnn_bias_regularizer,
                kernel_constraint=self.rnn_kernel_constraint,
                recurrent_constraint=self.rnn_recurrent_constraint,
                projection_constraint=self.rnn_projection_constraint,
                bias_constraint=self.rnn_bias_constraint,
                name="rnn",
            )(self.text2vec(token_ids)),
            mask,
            tag_ids,
        )

    def build_output_layer(self, inputs):
        embeddings, mask, tag_ids = inputs
        emissions = MLPLayer(
            self.num_fc_layers,
            hidden_size=self.fc_hidden_size,
            output_size=self.num_classes,
            activation=self.fc_activation,
            batch_normalization=self.use_batch_normalization,
            momentum=self.fc_momentum,
            epsilon=self.fc_epsilon,
            name="mlp",
        )(embeddings)
        return CrfLossLayer(
            self.num_classes, max_sequence_length=self.max_sequence_length
        )([emissions, tag_ids], mask)

    @classmethod
    def _filter_config(cls, config: Dict[str, Any]) -> Dict[str, Any]:
        config = super()._filter_config(config)
        config.pop("start_tag", None)
        config.pop("end_tag", None)
        return config

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "MLPLayer": MLPLayer,
            "LSTMPCell": LSTMPCell,
            "LSTMP": LSTMP,
            "MultiLSTMP": MultiLSTMP,
            "GlorotUniform": tf.keras.initializers.GlorotUniform,
            "Orthogonal": tf.keras.initializers.Orthogonal,
            "Zeros": tf.keras.initializers.Zeros,
        }


from typing import Sequence, List, Optional, Tuple, Any

import numpy as np
import pandas as pd
import tensorflow as tf

from sknlp.vocab import Vocab
from .nlp_dataset import NLPDataset


def _combine_xy(x, y):
    return (x, y), y


class TaggingDataset(NLPDataset):
    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        no_label: bool = False,
        start_tag: Optional[str] = None,
        end_tag: Optional[str] = None,
        max_length: Optional[int] = None,
        text_segmenter: str = "char",
        text_dtype: tf.DType = tf.int32,
        label_dtype: tf.DType = tf.int32,
    ):
        self.vocab = vocab
        self.start_tag = start_tag
        self.end_tag = end_tag
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(
            df=df,
            csv_file=csv_file,
            in_memory=in_memory,
            no_label=no_label,
            text_segmenter=text_segmenter,
            max_length=max_length,
            na_value="",
            column_dtypes=["str", "str"],
            text_dtype=text_dtype,
            label_dtype=label_dtype,
        )

    @property
    def y(self) -> List[List[str]]:
        if self.no_label:
            return []
        return [
            data[-1].decode("utf-8").split("|")
            for data in self._original_dataset.as_numpy_iterator()
        ]

    @property
    def batch_padding_shapes(self) -> List[Tuple]:
        return ((None,), (None,))

    def _text_transform(self, text: tf.Tensor) -> np.ndarray:
        tokens = super()._text_transform(text)
        return np.array(
            [self.vocab[t] for t in tokens[: self.max_length]], dtype=np.int32
        )

    def _label_transform(self, label: tf.Tensor) -> List[int]:
        label = super()._label_transform(label)
        labels = [self.label2idx[l] for l in label.split("|")][: self.max_length]
        if self.start_tag is not None and self.end_tag is not None:
            labels = [
                self.label2idx[self.start_tag],
                *labels,
                self.label2idx[self.end_tag],
            ]
        return labels

    def _transform_func(self, *data) -> List[Any]:
        text = data[0]
        if self.no_label:
            _text = self._text_transform(text)
            return _text, [0 for _ in range(len(_text))]
        label = data[1]
        return self._text_transform(text), self._label_transform(label)

    def _transform_func_out_dtype(self) -> List[tf.DType]:
        return (self.text_dtype, self.label_dtype)

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        return super().batchify(
            batch_size,
            shuffle=shuffle,
            shuffle_buffer_size=shuffle_buffer_size,
            after_batch=_combine_xy,
        )



from typing import Type, Sequence, Dict, Any, List, Optional, Union
import os
import contextlib

import tensorflow as tf
import kerastuner as kt

from sknlp.callbacks import default_supervised_model_callbacks
from sknlp.data import NLPDataset
from sknlp.layers import BertLayer
from .text2vec import Text2vec
from .supervised_model import SupervisedNLPModel


@contextlib.contextmanager
def maybe_distribute(distribution_strategy):
    """Distributes if distribution_strategy is set."""
    if distribution_strategy is None:
        yield
    else:
        with distribution_strategy.scope():
            yield


def create_model_builder(
    model_type: Type[SupervisedNLPModel],
    model_args: List[Any],
    model_kwargs: Dict[str, Any],
    optimizer: str,
    optimizer_parameters: Dict[str, Any],
    distribute_strategy: Optional[tf.distribute.Strategy] = None,
):
    text2vec: Text2vec = model_kwargs.pop("text2vec", None)

    def model_builder(hp):
        with maybe_distribute(distribute_strategy):
            clone_text2vec: Optional[Text2vec] = None
            if text2vec is not None:
                clone_text2vec: Text2vec = Text2vec(
                    text2vec.vocab,
                    segmenter=text2vec.segmenter,
                    max_sequence_length=text2vec.max_sequence_length,
                    sequence_length=text2vec.sequence_length,
                    embedding_size=text2vec.embedding_size,
                )
                keras_model: tf.keras.Model = tf.keras.models.model_from_json(
                    text2vec._model.to_json(),
                    custom_objects={
                        "TruncatedNormal": tf.keras.initializers.TruncatedNormal,
                        "BertLayer": BertLayer,
                    },
                )
                keras_model.set_weights(text2vec._model.get_weights())
                clone_text2vec._model = keras_model
                if not text2vec._model.layers[0].trainable:
                    clone_text2vec.freeze()
            model = model_type(*model_args, text2vec=clone_text2vec, **model_kwargs)
            model.compile_optimizer(optimizer, **optimizer_parameters)
            return model._model

    return model_builder


class ParameterSearcher:
    def __init__(
        self,
        model_type: Type[SupervisedNLPModel],
        hyper_parameters: kt.HyperParameters,
        *args,
        **kwargs,
    ) -> None:
        self.model_type = model_type
        self.hyper_parameters = hyper_parameters
        self.model_args = args
        self.model_kwargs = kwargs
        self.temp_model: SupervisedNLPModel = model_type(*args, **kwargs)
        self.tuner = None

    def search(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        validation_X: Sequence[str] = None,
        validation_y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        validation_dataset: NLPDataset = None,
        batch_size: int = 128,
        n_epochs: int = 10,
        optimizer: str = "adam",
        optimizer_kwargs: Optional[Dict[str, Any]] = None,
        learning_rate: float = 1e-3,
        weight_decay: float = 0.0,
        clip: Optional[float] = 5.0,
        learning_rate_update_factor: float = 0.5,
        learning_rate_update_epochs: int = 10,
        enable_early_stopping: bool = False,
        early_stopping_patience: Optional[int] = None,
        early_stopping_min_delta: float = 0.0,
        early_stopping_use_best_epoch: bool = False,
        early_stopping_monitor: int = 2,  # 1 for loss, 2 for metric
        checkpoint: Optional[str] = None,
        save_frequency: int = 1,
        log_file: Optional[str] = None,
        verbose: int = 2,
        distribute_strategy: Optional[tf.distribute.Strategy] = None,
        max_trials: int = 20,
        executions_per_trial: int = 3,
        search_result_directory: Optional[str] = None,
    ) -> None:
        training_dataset = self.temp_model.prepare_dataset(X, y, dataset)
        assert (
            validation_X is None or validation_y is None
        ) or validation_dataset is None, "No validation set is provided."
        validation_dataset = self.temp_model.prepare_dataset(
            validation_X, validation_y, validation_dataset
        )
        training_tf_dataset = training_dataset.batchify(batch_size)
        validation_tf_dataset = validation_dataset.batchify(batch_size, shuffle=False)

        monitor = "val_loss"
        monitor_direction = "min"
        if early_stopping_monitor == 2 and self.temp_model.get_monitor():
            monitor = self.temp_model.get_monitor()
            monitor_direction = "max"
        callbacks = default_supervised_model_callbacks(
            learning_rate_update_factor=learning_rate_update_factor,
            learning_rate_update_epochs=learning_rate_update_epochs,
            use_weight_decay=weight_decay > 0,
            enable_early_stopping=enable_early_stopping,
            early_stopping_monitor=monitor,
            early_stopping_monitor_direction=monitor_direction,
            early_stopping_patience=early_stopping_patience,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_use_best_epoch=early_stopping_use_best_epoch,
            log_file=log_file,
        )

        optimizer_kwargs = optimizer_kwargs or dict()
        optimizer_parameters = {
            "learning_rate": learning_rate,
            "weight_decay": weight_decay,
            **optimizer_kwargs,
        }
        if clip is not None:
            optimizer_parameters["clipnorm"] = clip
        model_builder = create_model_builder(
            self.model_type,
            self.model_args,
            self.model_kwargs,
            optimizer,
            optimizer_parameters,
            distribute_strategy=distribute_strategy,
        )

        objective = kt.Objective(monitor, monitor_direction)
        directory = None
        project_name = None
        if search_result_directory is not None:
            search_result_directory = search_result_directory.rstrip(".\\/")
            directory = os.path.dirname(search_result_directory)
            project_name = os.path.basename(search_result_directory)
            if directory == project_name:
                project_name = "default"
        self.tuner = kt.RandomSearch(
            model_builder,
            objective=objective,
            hyperparameters=self.hyper_parameters,
            max_trials=max_trials,
            executions_per_trial=executions_per_trial,
            directory=directory,
            project_name=project_name,
            distribution_strategy=distribute_strategy,
            overwrite=False,
        )
        self.tuner.search(
            training_tf_dataset,
            epochs=n_epochs,
            validation_data=validation_tf_dataset,
            callbacks=callbacks,
            verbose=verbose,
        )

    def summary(self) -> Dict[str, Any]:
        if self.tuner is not None:
            return self.tuner.results_summary()
        return {}



from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as K


class LearningRateScheduler(Callback):
    def __init__(self, warmup_schedule, decay_schedule, verbose=0):
        super().__init__()
        self.warmup_schedule = warmup_schedule
        self.decay_schedule = decay_schedule
        self.epoch = 0
        self.verbose = verbose

    def on_epoch_begin(self, epoch, logs):
        self.epoch = epoch
        if not hasattr(self.model.optimizer, "lr"):
            raise ValueError('Optimizer must have a "lr" attribute.')
        lr = float(K.get_value(self.model.optimizer.lr))
        new_lr = self.decay_schedule(epoch, lr)
        if new_lr == lr:
            return
        K.set_value(self.model.optimizer.lr, K.get_value(new_lr))
        if self.verbose > 0:
            print(
                "\nEpoch %05d: LearningRateScheduler decreasing learning rate "
                "to %s." % (epoch + 1, new_lr)
            )

    def on_train_batch_begin(self, batch, logs):
        if not hasattr(self.model.optimizer, "lr"):
            raise ValueError('Optimizer must have a "lr" attribute.')
        lr = float(K.get_value(self.model.optimizer.lr))
        step = self.epoch * self.params["steps"] + batch
        new_lr = self.warmup_schedule(step, lr)
        if new_lr == lr:
            return
        K.set_value(self.model.optimizer.lr, K.get_value(new_lr))
        if self.verbose > 0:
            print(
                "\nStep %07d: LearningRateScheduler increasing learning rate "
                "to %s." % (step + 1, new_lr)
            )

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        logs["lr"] = K.get_value(self.model.optimizer.lr)


from __future__ import annotations
from typing import Optional, Sequence, Callable
from dataclasses import dataclass
import itertools

import numpy as np
import pandas as pd


@dataclass
class Tag:
    label: str
    start: int
    end: str

    def __hash__(self) -> int:
        return hash((self.label, self.start, self.end))


def parse_tagged_text(tag_names: Sequence[str]) -> list[Tag]:
    current_label = None
    start = 0
    parsed_tags = list()
    for i, tag_name in enumerate(tag_names):
        if tag_name == "O":
            tag_type, tag_label = "O", None
        else:
            tag_type, tag_label = tag_name.split("-")
        if (tag_type == "I" or tag_type == "E") and tag_label == current_label:
            continue

        if i != start:
            parsed_tags.append(Tag(current_label, start, i))

        if tag_type == "B" or tag_type == "S":
            start = i
            current_label = tag_label
        else:
            start = i + 1
            current_label = None
    if start != len(tag_names) and current_label is not None:
        parsed_tags.append(Tag(current_label, start, len(tag_names)))
    return parsed_tags


def _compute_counts(
    label: Sequence[str], prediction: Sequence[str], classes: Sequence[str]
) -> list[tuple[str, int, int, int]]:
    truth_tags = set(parse_tagged_text(label))
    prediction_tags = set(parse_tagged_text(prediction))
    correct_tags = truth_tags & prediction_tags
    counts = [
        (
            cls,
            len([i for i, tag in enumerate(correct_tags) if tag.label == cls]),
            len([i for i, tag in enumerate(truth_tags) if tag.label == cls]),
            len([i for i, tag in enumerate(prediction_tags) if tag.label == cls]),
        )
        for cls in classes
    ]
    counts.append(("avg", len(correct_tags), len(truth_tags), len(prediction_tags)))
    return counts


def tagging_fscore(
    y: Sequence[Sequence[str]],
    p: Sequence[Sequence[str]],
    classes: Sequence[str],
) -> pd.DataFrame:
    df = pd.DataFrame(
        itertools.chain.from_iterable(
            _compute_counts(yi, pi, classes) for yi, pi in zip(y, p)
        ),
        columns=["class", "correct", "support", "prediction"],
    )
    df = df.groupby("class").sum()
    df["precision"] = df.correct / df.prediction
    df["recall"] = df.correct / df.support
    df["fscore"] = 2 * df.precision * df.recall / (df.precision + df.recall)
    df.fillna(0, inplace=True)
    df.reset_index(inplace=True)
    return df[["class", "precision", "recall", "fscore", "support"]]


def convert_ids_to_tags(
    idx2tag: Callable[[int], str],
    tag_ids: list[int],
    pad_tag: str,
    start_tag: Optional[str] = None,
    end_tag: Optional[str] = None,
) -> list[str]:
    exclude_tags = {pad_tag, start_tag, end_tag}
    tags = []
    for tag_id in tag_ids:
        tag = idx2tag(tag_id)
        if tag not in exclude_tags:
            tags.append(tag)
        if tag == pad_tag:
            break
    return tags


def viterbi_decode(transitions, emissions, mask=None):
    """
    Decode the highest scoring sequence of tags outside of tensorflow.
    Parameters:
    ----
    transitions: numpy array, shape(num_tags, num_tags)
    emissions: numpy array, shape(seq_length, batch_size, num_tags)
    mask: numpy array, shape(seq_length, batch_size)
    Returns:
    ----
    best_tags_list: list of list
      Each list in best_tags_list is the best path of each sequence.
    """
    seq_legnth, batch_size, num_tags = emissions.shape
    if mask is None:
        mask = np.ones((seq_legnth, batch_size))
    assert mask[0].all()

    sequence_lengths = mask.sum(axis=0).astype(np.int64)
    # list to store the decode paths
    best_tags_list = []
    # (seq_length, batch_size, num_tags)
    viterbi_scores = np.zeros_like(emissions, dtype=np.float32)
    viterbi_scores[0] = emissions[0]
    # (seq_length, batch_size, num_tags)
    viterbi_paths = np.zeros_like(emissions, dtype=np.int64)

    # use dynamic programing to compute the viterbi score
    for i, emission in enumerate(emissions[1:]):
        # (batch_size, num_tags, 1)
        broadcast_log_prob = viterbi_scores[i].reshape((-1, num_tags, 1))
        # (batch_size, num_tags, num_tags)
        score = broadcast_log_prob + transitions
        viterbi_paths[i] = score.argmax(axis=1)
        viterbi_scores[i + 1] = emission + score.max(axis=1)
    # search best path for each batch according to the viterbi score
    # get the best tag for the last emission
    for idx in range(batch_size):
        seq_end_idx = sequence_lengths[idx] - 1
        best_last_tag = viterbi_scores[seq_end_idx, idx].argmax()
        best_tags = [best_last_tag]
        # trace back all best tags based on the last best tag and viterbi path
        for path in np.flip(viterbi_paths[: sequence_lengths[idx] - 1], 0):
            best_last_tag = path[idx][best_tags[-1]]
            best_tags.append(best_last_tag)
        best_tags.reverse()
        best_tags_list.append(best_tags)
    return best_tags_list


from functools import partial

import tensorflow as tf
from tensorflow.python.keras.losses import LossFunctionWrapper
import tensorflow.keras.backend as K
from sknlp.typing import WeightInitializer


def soft_fscore(activation_function, y_true, y_pred, from_logits=False, label_smoothing=0):
    if from_logits:
        y_pred = activation_function(y_pred)
    tp = K.sum(y_true * y_pred, axis=0)
    fp = K.sum((1 - y_true) * y_pred, axis=0)
    tn = K.sum((1 - y_true) * (1 - y_pred), axis=0)
    fn = K.sum(y_true * (1 - y_pred), axis=0)
    return (
        1
        - tp / (2 * tp + fn + fp + K.epsilon())
        - tn / (2 * tn + fn + fp + K.epsilon())
    )


class SoftFScoreLoss(LossFunctionWrapper):
    def __init__(self,
                 is_multilabel=False,
                 from_logits=False,
                 label_smoothing=0,
                 reduction=tf.keras.losses.Reduction.NONE,
                 name="soft-f-score"):
        func = K.softmax
        if is_multilabel:
            func = K.sigmoid
        super().__init__(
            partial(soft_fscore, func),
            name=name,
            reduction=reduction,
            from_logits=from_logits,
            label_smoothing=label_smoothing)
        self.from_logits = from_logits


from tensorflow.keras.layers import Layer


class CrfLosts(Layer):

    def __init__(
        self,
        output_size: int = 1,
        kernel_initializer: WeightInitializer = 'glorot_uniform',
        name: str = "crf",
        **kwargs
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.output_size = output_size

    def build(self, input_shape):
        self.kernel = self.add_weight(
            shape=(self.output_size, self.output_size),
            name='kernel',
            initializer=self.kernel_initializer,
        )



import logging
from functools import partial

import tensorflow as tf
from official.modeling import activations
from official.nlp.modeling import networks

from .bert_converter import convert_checkpoint, BERT_PERMUTATIONS


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
logger.addHandler(handler)
logger.setLevel(logging.INFO)


ALBERT_NAME_REPLACEMENTS = (
    ("encoder/", ""),
    ("embeddings/word_embeddings", "word_embeddings/embeddings"),
    ("embeddings/position_embeddings", "position_embedding/embeddings"),
    ("embeddings/token_type_embeddings", "type_embeddings/embeddings"),
    ("embeddings/LayerNorm", "embeddings/layer_norm"),
    ("embedding_hidden_mapping_in", "embedding_projection"),
    ("group_0/inner_group_0/", ""),
    ("attention_1/self", "self_attention"),
    ("attention_1/output/dense", "self_attention/attention_output"),
    ("LayerNorm/", "self_attention_layer_norm/"),
    ("ffn_1/intermediate/dense", "intermediate"),
    ("ffn_1/intermediate/output/dense", "output"),
    ("LayerNorm_1/", "output_layer_norm/"),
    ("pooler/dense", "pooler_transform"),
    ("cls/predictions/output_bias", "cls/predictions/output_bias/bias"),
    ("cls/seq_relationship/output_bias", "predictions/transform/logits/bias"),
    ("cls/seq_relationship/output_weights", "predictions/transform/logits/kernel"),
)


convert_albert_checkpoint = partial(
    convert_checkpoint,
    name_replacements=ALBERT_NAME_REPLACEMENTS,
    permutations=BERT_PERMUTATIONS,
)



import functools

try:
    import jieba_fast as jieba
except ImportError:
    import jieba


def get_segmenter(name):
    if name == "jieba":
        return functools.partial(jieba.lcut, HMM=False)
    elif name == "list" or name == "char":
        return list
    elif name is None:
        return lambda x: x
    else:
        raise ValueError("unknown segmenter %s" % name)



from .bert_converter import convert_bert_checkpoint, convert_electra_checkpoint
from .albert_converter import convert_albert_checkpoint


__all__ = [
    "convert_bert_checkpoint",
    "convert_electra_checkpoint",
    "convert_albert_checkpoint",
]



from .precision_recall_f1 import (
    PrecisionWithLogits,
    RecallWithLogits,
    FBetaScoreWithLogits,
)
from .accuracy import BinaryAccuracyWithLogits, AccuracyWithLogits


__all__ = [
    "BinaryAccuracyWithLogits",
    "AccuracyWithLogits",
    "PrecisionWithLogits",
    "RecallWithLogits",
    "FBetaScoreWithLogits",
]



from typing import List, Dict, Any, Optional

import tensorflow as tf
from tensorflow_addons.text import crf_log_likelihood, crf_decode

from sknlp.typing import WeightInitializer


class CrfLossLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        num_tags: int,
        max_sequence_length: int = 120,
        initializer: WeightInitializer = "Orthogonal",
        name: str = "crf",
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.num_tags = num_tags
        self.max_sequence_length = max_sequence_length
        self.initializer = initializer
        self.transition_weight = self.add_weight(
            shape=(self.num_tags, self.num_tags),
            dtype=tf.float32,
            initializer=tf.keras.initializers.get(initializer),
            name="transision_weight",
        )

    def call(self, inputs: List[tf.Tensor], mask: tf.Tensor) -> tf.Tensor:
        emissions, tag_ids = inputs
        mask = tf.cast(mask, tf.int32)
        sequence_lengths = tf.math.reduce_sum(mask, axis=1)
        likelihoods, _ = crf_log_likelihood(
            emissions, tag_ids, sequence_lengths, self.transition_weight
        )
        loss = tf.math.negative(tf.math.reduce_mean(likelihoods))
        self.add_loss(loss)
        return (
            tf.pad(
                emissions,
                [[0, 0], [0, self.max_sequence_length + 2 - tf.shape(mask)[1]], [0, 0]],
            ),
            tf.pad(
                mask, [[0, 0], [0, self.max_sequence_length + 2 - tf.shape(mask)[1]]]
            ),
        )

    def get_config(self):
        base_config = super().get_config()
        return {
            **base_config,
            "num_tags": self.num_tags,
            "max_sequence_length": self.max_sequence_length,
            "initializer": self.initializer,
        }


class CrfDecodeLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        num_tags: int,
        max_sequence_length: int = 120,
        initializer: WeightInitializer = "Orthogonal",
        name: str = "crf",
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.num_tags = num_tags
        self.max_sequence_length = max_sequence_length
        self.initializer = initializer
        self.transition_weight = self.add_weight(
            shape=(self.num_tags, self.num_tags),
            dtype=tf.float32,
            initializer=tf.keras.initializers.get(initializer),
            name="transision_weight",
        )

    def call(self, emissions: tf.Tensor, mask: tf.Tensor) -> tf.Tensor:
        sequence_lengths = tf.math.reduce_sum(mask, axis=1)
        decoded_tag_ids, _ = crf_decode(
            emissions, self.transition_weight, sequence_lengths
        )
        return tf.pad(
            decoded_tag_ids * mask,
            [[0, 0], [0, self.max_sequence_length + 2 - tf.shape(decoded_tag_ids)[1]]],
            mode="CONSTANT",
        )

    def get_config(self):
        base_config = super().get_config()
        return {
            **base_config,
            "num_tags": self.num_tags,
            "max_sequence_length": self.max_sequence_length,
            "initializer": self.initializer,
        }



from typing import Sequence, List, Dict, Any, Optional, Callable

import tensorflow as tf
from tensorflow.keras.activations import gelu
from tensorflow.keras.initializers import TruncatedNormal
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

from official.modeling import activations
from official.nlp.keras_nlp import layers
import tensorflow_text as tftext

from .bert_tokenization import BertTokenizationLayer


class BertCharPreprocessingLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        vocab: Sequence[str],
        cls_token: str = "[CLS]",
        sep_token: str = "[SEP]",
        **kwargs
    ) -> None:
        vocab: List = list(vocab)
        self.cls_token = cls_token
        self.sep_token = sep_token
        if cls_token not in vocab:
            vocab.append(cls_token)
        if sep_token not in vocab:
            vocab.append(sep_token)
        if "[UNK]" in vocab:
            idx = vocab.index("[UNK]")
            vocab[idx] = "unk"
        self.vocab = vocab
        self.tokenizer = TextVectorization(
            max_tokens=len(vocab),
            standardize=None,
            split=BertTokenizationLayer(cls_token=cls_token, sep_token=sep_token),
        )
        self.tokenizer.set_vocabulary(vocab[2:])
        super().__init__(**kwargs)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self.tokenizer(inputs)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "vocab": self.vocab,
            "cls_token": self.cls_token,
            "sep_token": self.sep_token,
        }


class BertPreprocessingLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        vocab: Sequence[str],
        cls_token: str = "[CLS]",
        sep_token: str = "[SEP]",
        **kwargs
    ) -> None:
        vocab: List = list(vocab)
        self.cls_token = cls_token
        self.sep_token = sep_token
        if cls_token not in vocab:
            vocab.append(cls_token)
        if sep_token not in vocab:
            vocab.append(sep_token)
        self.vocab = vocab
        self.tokenizer = tftext.BertTokenizer(
            tf.lookup.StaticVocabularyTable(
                tf.lookup.KeyValueTensorInitializer(
                    vocab,
                    list(range(len(vocab))),
                    key_dtype=tf.string,
                    value_dtype=tf.int64,
                ),
                1,
            )
        )
        for i, token in enumerate(vocab):
            if token == cls_token:
                self.cls_id = i
            if token == sep_token:
                self.sep_id = i
        super().__init__(**kwargs)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        token_ids = self.tokenizer.tokenize(inputs)
        flatten_ids = token_ids.merge_dims(-2, -1)
        cls_ids = tf.reshape(
            tf.tile(tf.constant([self.cls_id], dtype=tf.int64), [token_ids.nrows()]),
            [token_ids.nrows(), 1],
        )
        sep_ids = tf.reshape(
            tf.tile(tf.constant([self.sep_id], dtype=tf.int64), [token_ids.nrows()]),
            [token_ids.nrows(), 1],
        )
        ids = tf.concat([cls_ids, flatten_ids, sep_ids], axis=1)
        return ids.to_tensor(0)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "vocab": self.vocab,
            "cls_token": self.cls_token,
            "sep_token": self.sep_token,
        }


class BertPairPreprocessingLayer(BertPreprocessingLayer):
    def call(self, inputs: List[tf.Tensor]) -> List[tf.Tensor]:
        query, context = inputs
        query_token_ids = self.tokenizer.tokenize(query)
        context_token_ids = self.tokenizer.tokenize(context)
        query_flatten_ids = query_token_ids.merge_dims(-2, -1)
        context_flatten_ids = context_token_ids.merge_dims(-2, -1)
        cls_ids = tf.reshape(
            tf.tile(
                tf.constant([self.cls_id], dtype=tf.int64), [query_token_ids.nrows()]
            ),
            [query_token_ids.nrows(), 1],
        )
        sep_ids = tf.reshape(
            tf.tile(
                tf.constant([self.sep_id], dtype=tf.int64), [query_token_ids.nrows()]
            ),
            [query_token_ids.nrows(), 1],
        )
        query_ids = tf.concat([cls_ids, query_flatten_ids, sep_ids], axis=1)
        context_ids = tf.concat([context_flatten_ids, sep_ids], axis=1)
        type_ids = tf.concat(
            [
                tf.zeros_like(query_ids, dtype=tf.int64),
                tf.ones_like(context_ids, dtype=tf.int64),
            ],
            axis=1,
        )
        return (
            tf.concat([query_ids, context_ids], axis=1, name="token_ids").to_tensor(0),
            type_ids.to_tensor(0),
        )


class BertLayer(tf.keras.layers.Layer):
    """
    copy from https://github.com/tensorflow/models/blob/v2.2.0/official/nlp/modeling/networks/transformer_encoder.py.
    """

    def __init__(
        self,
        vocab_size: int = 100,
        embedding_size: int = 768,
        hidden_size: int = 768,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        sequence_length: int = None,
        max_sequence_length: int = 512,
        type_vocab_size: int = 16,
        intermediate_size: int = 3072,
        activation: Callable = gelu,
        dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        initializer: tf.keras.initializers.Initializer = TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs: bool = False,
        share_layer: bool = False,
        cls_pooling: bool = True,
        name: str = "bert_layer",
        **kwargs
    ) -> None:
        if embedding_size is None:
            embedding_size = hidden_size

        self.vocab_size = vocab_size
        self.embedding_size = embedding_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_attention_heads = num_attention_heads
        self.sequence_length = sequence_length
        self.max_sequence_length = max_sequence_length
        self.type_vocab_size = type_vocab_size
        self.intermediate_size = intermediate_size
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.attention_dropout_rate = attention_dropout_rate
        self.initializer = initializer
        self.return_all_encoder_outputs = return_all_encoder_outputs
        self.share_layer = share_layer
        self.cls_pooling = cls_pooling

        self.embedding_layer = layers.OnDeviceEmbedding(
            vocab_size, embedding_size, initializer=initializer, name="word_embeddings"
        )
        # Always uses dynamic slicing for simplicity.
        self.position_embedding_layer = layers.PositionEmbedding(
            max_sequence_length,
            initializer=initializer,
            name="position_embedding",
        )
        self.type_embedding_layer = layers.OnDeviceEmbedding(
            type_vocab_size,
            embedding_size,
            initializer=initializer,
            use_one_hot=True,
            name="type_embeddings",
        )
        if embedding_size != hidden_size:
            self.embedding_projection = tf.keras.layers.experimental.EinsumDense(
                "...x,xy->...y",
                output_shape=hidden_size,
                bias_axes="y",
                kernel_initializer=initializer,
                name="embedding_projection",
            )
        else:
            self.embedding_projection = None
        self.normalize_layer = tf.keras.layers.LayerNormalization(
            name="embeddings/layer_norm", axis=-1, epsilon=1e-12, dtype=tf.float32
        )
        if share_layer:
            self.shared_layer = layers.TransformerEncoderBlock(
                num_attention_heads,
                intermediate_size,
                activation,
                output_dropout=dropout_rate,
                attention_dropout=attention_dropout_rate,
                kernel_initializer=initializer,
                name="transformer",
            )
        else:
            self.transformer_layers = []
            for i in range(num_layers):
                layer = layers.TransformerEncoderBlock(
                    num_attention_heads,
                    intermediate_size,
                    activation,
                    output_dropout=dropout_rate,
                    attention_dropout=attention_dropout_rate,
                    kernel_initializer=initializer,
                    name="transformer/layer_%d" % i,
                )
                self.transformer_layers.append(layer)

        if cls_pooling:
            self.cls_output_layer = tf.keras.layers.Dense(
                units=hidden_size,
                activation="tanh",
                kernel_initializer=initializer,
                name="pooler_transform",
            )
        super().__init__(name=name, **kwargs)

    def call(
        self, inputs: List[tf.Tensor], mask: Optional[tf.Tensor] = None
    ) -> tf.Tensor:
        token_ids, type_ids = inputs
        mask = self.compute_mask(inputs, mask=mask)
        word_embeddings = self.embedding_layer(token_ids)
        position_embeddings = self.position_embedding_layer(word_embeddings)
        type_embeddings = self.type_embedding_layer(type_ids)
        embeddings = tf.keras.layers.Add()(
            [word_embeddings, position_embeddings, type_embeddings]
        )
        embeddings = self.normalize_layer(embeddings)
        embeddings = tf.keras.layers.Dropout(rate=self.dropout_rate)(embeddings)
        if self.embedding_projection is not None:
            embeddings = self.embedding_projection(embeddings)
        data = embeddings
        attention_mask = layers.SelfAttentionMask()(data, mask)
        encoder_outputs = []
        if self.share_layer:
            for _ in range(self.num_layers):
                data = self.shared_layer([data, attention_mask])
                encoder_outputs.append(data)
        else:
            for layer in self.transformer_layers:
                data = layer([data, attention_mask])
                encoder_outputs.append(data)

        first_token_tensor = tf.keras.layers.Lambda(
            lambda x: tf.squeeze(x[:, 0:1, :], axis=1)
        )(encoder_outputs[-1])
        cls_output = first_token_tensor
        if self.cls_pooling:
            cls_output = self.cls_output_layer(cls_output)
        if self.return_all_encoder_outputs:
            return [encoder_outputs, mask, cls_output]
        else:
            return [encoder_outputs[-1], mask, cls_output]

    def compute_mask(
        self, inputs: List[tf.Tensor], mask: Optional[tf.Tensor] = None
    ) -> tf.Tensor:
        if mask is not None:
            return mask
        return tf.math.not_equal(inputs[0], 0)

    def get_config(self):
        return {
            **super().get_config(),
            "vocab_size": self.vocab_size,
            "embedding_size": self.embedding_size,
            "hidden_size": self.hidden_size,
            "num_layers": self.num_layers,
            "num_attention_heads": self.num_attention_heads,
            "sequence_length": self.sequence_length,
            "max_sequence_length": self.max_sequence_length,
            "type_vocab_size": self.type_vocab_size,
            "intermediate_size": self.intermediate_size,
            "activation": self.activation,
            "dropout_rate": self.dropout_rate,
            "attention_dropout_rate": self.attention_dropout_rate,
            "initializer": self.initializer,
            "return_all_encoder_outputs": self.return_all_encoder_outputs,
            "share_layer": self.share_layer,
            "cls_pooling": self.cls_pooling,
        }